@@ -15,10 +15,10 @@ from collections.abc import (
 from contextlib import ContextDecorator, contextmanager, suppress
 from dataclasses import dataclass, field
 from enum import Enum, auto
-from functools import singledispatchmethod, wraps
-from inspect import Signature, get_annotations
+from functools import partialmethod, singledispatchmethod, wraps
+from inspect import Signature, isclass
 from threading import RLock
-from types import MappingProxyType
+from types import MappingProxyType, UnionType
 from typing import (
     Any,
     ContextManager,
@@ -26,13 +26,12 @@ from typing import (
     Protocol,
     TypeVar,
     cast,
-    final,
     runtime_checkable,
 )
 
 from injection.common.event import Event, EventChannel, EventListener
 from injection.common.lazy import Lazy, LazyMapping
-from injection.common.tools import format_type, get_origin
+from injection.common.tools import find_types, format_type, get_origins
 from injection.exceptions import (
     ModuleError,
     ModuleLockError,
@@ -46,6 +45,7 @@ _logger = logging.getLogger(__name__)
 _thread_lock = RLock()
 
 _T = TypeVar("_T")
+Types = Iterable[type] | UnionType
 
 
 """
@@ -132,6 +132,9 @@ Injectables
 class Injectable(Protocol[_T]):
     __slots__ = ()
 
+    def __init__(self, factory: Callable[[], _T] = ..., *args, **kwargs):
+        ...
+
     @property
     def is_locked(self) -> bool:
         return False
@@ -193,16 +196,15 @@ class Container:
     __data: dict[type, Injectable] = field(default_factory=dict, init=False)
     __channel: EventChannel = field(default_factory=EventChannel, init=False)
 
-    def __getitem__(self, cls: type[_T], /) -> Injectable[_T]:
-        origin = get_origin(cls)
+    def __getitem__(self, cls: type[_T] | UnionType, /) -> Injectable[_T]:
+        for origin in get_origins(cls):
+            with suppress(KeyError):
+                return self.__data[origin]
 
-        try:
-            return self.__data[origin]
-        except KeyError as exc:
-            raise NoInjectable(cls) from exc
+        raise NoInjectable(cls)
 
-    def __contains__(self, cls: type, /) -> bool:
-        return get_origin(cls) in self.__data
+    def __contains__(self, cls: type | UnionType, /) -> bool:
+        return any(origin in self.__data for origin in get_origins(cls))
 
     @property
     def is_locked(self) -> bool:
@@ -212,8 +214,8 @@ class Container:
     def __injectables(self) -> frozenset[Injectable]:
         return frozenset(self.__data.values())
 
-    def update(self, classes: Iterable[type], injectable: Injectable):
-        classes = frozenset(get_origin(cls) for cls in classes)
+    def update(self, classes: Types, injectable: Injectable):
+        classes = frozenset(get_origins(*classes))
 
         if classes:
             event = ContainerDependenciesUpdated(self, classes)
@@ -272,34 +274,22 @@ class Module(EventListener):
     def __post_init__(self):
         self.__container.add_listener(self)
 
-    def __getitem__(self, cls: type[_T], /) -> Injectable[_T]:
+    def __getitem__(self, cls: type[_T] | UnionType, /) -> Injectable[_T]:
         for broker in self.__brokers:
             with suppress(KeyError):
                 return broker[cls]
 
         raise NoInjectable(cls)
 
-    def __setitem__(self, cls: type, injectable: Injectable, /):
+    def __setitem__(self, cls: type | UnionType, injectable: Injectable, /):
         self.update((cls,), injectable)
 
-    def __contains__(self, cls: type, /) -> bool:
+    def __contains__(self, cls: type | UnionType, /) -> bool:
         return any(cls in broker for broker in self.__brokers)
 
     def __str__(self) -> str:
         return self.name or object.__str__(self)
 
-    @property
-    def inject(self) -> InjectDecorator:
-        return InjectDecorator(self)
-
-    @property
-    def injectable(self) -> InjectableDecorator:
-        return InjectableDecorator(self, NewInjectable)
-
-    @property
-    def singleton(self) -> InjectableDecorator:
-        return InjectableDecorator(self, SingletonInjectable)
-
     @property
     def is_locked(self) -> bool:
         return any(broker.is_locked for broker in self.__brokers)
@@ -309,6 +299,57 @@ class Module(EventListener):
         yield from tuple(self.__modules)
         yield self.__container
 
+    def injectable(
+        self,
+        wrapped: Callable[..., Any] = None,
+        /,
+        *,
+        cls: type[Injectable] = NewInjectable,
+        on: type | Types = None,
+    ):
+        def decorator(wp):
+            factory = self.inject(wp, return_factory=True)
+            injectable = cls(factory)
+            classes = find_types(wp, on)
+            self.update(classes, injectable)
+            return wp
+
+        return decorator(wrapped) if wrapped else decorator
+
+    singleton = partialmethod(injectable, cls=SingletonInjectable)
+
+    def set_constant(self, instance: _T, on: type | Types = None) -> _T:
+        cls = type(instance)
+
+        @self.injectable(on=(cls, on))
+        def get_constant():
+            return instance
+
+        return instance
+
+    def inject(
+        self,
+        wrapped: Callable[..., Any] = None,
+        /,
+        *,
+        return_factory: bool = False,
+    ):
+        def decorator(wp):
+            if not return_factory and isclass(wp):
+                wp.__init__ = decorator(wp.__init__)
+                return wp
+
+            lazy_binder = Lazy[Binder](lambda: self.__new_binder(wp))
+
+            @wraps(wp)
+            def wrapper(*args, **kwargs):
+                arguments = (~lazy_binder).bind(*args, **kwargs)
+                return wp(*arguments.args, **arguments.kwargs)
+
+            return wrapper
+
+        return decorator(wrapped) if wrapped else decorator
+
     def get_instance(self, cls: type[_T]) -> _T | None:
         try:
             injectable = self[cls]
@@ -321,7 +362,7 @@ class Module(EventListener):
     def get_lazy_instance(self, cls: type[_T]) -> Lazy[_T | None]:
         return Lazy(lambda: self.get_instance(cls))
 
-    def update(self, classes: Iterable[type], injectable: Injectable):
+    def update(self, classes: Types, injectable: Injectable):
         self.__container.update(classes, injectable)
         return self
 
@@ -395,7 +436,7 @@ class Module(EventListener):
 
         with self.__channel.dispatch(event):
             yield
-            _logger.debug(f"{event}")
+            _logger.debug(event)
 
     def __check_locking(self):
         if self.is_locked:
@@ -411,6 +452,12 @@ class Module(EventListener):
                 f"`{module}` can't be found in the modules used by `{self}`."
             ) from exc
 
+    def __new_binder(self, target: Callable[..., Any]) -> Binder:
+        signature = inspect.signature(target, eval_str=True)
+        binder = Binder(signature).update(self)
+        self.add_listener(binder)
+        return binder
+
 
 """
 Binder
@@ -493,98 +540,3 @@ class Binder(EventListener):
     def _(self, event: ModuleEvent, /) -> ContextManager:
         yield
         self.update(event.on_module)
-
-
-"""
-Decorators
-"""
-
-
-@final
-@dataclass(repr=False, frozen=True, slots=True)
-class InjectDecorator:
-    __module: Module
-
-    def __call__(self, wrapped: Callable[..., Any] = None, /):
-        def decorator(wp):
-            if isinstance(wp, type):
-                return self.__class_decorator(wp)
-
-            return self.__decorator(wp)
-
-        return decorator(wrapped) if wrapped else decorator
-
-    def __decorator(self, function: Callable[..., Any], /) -> Callable[..., Any]:
-        lazy_binder = Lazy[Binder](lambda: self.__new_binder(function))
-
-        @wraps(function)
-        def wrapper(*args, **kwargs):
-            arguments = (~lazy_binder).bind(*args, **kwargs)
-            return function(*arguments.args, **arguments.kwargs)
-
-        return wrapper
-
-    def __class_decorator(self, cls: type, /) -> type:
-        cls.__init__ = self.__decorator(cls.__init__)
-        return cls
-
-    def __new_binder(self, function: Callable[..., Any]) -> Binder:
-        signature = inspect.signature(function, eval_str=True)
-        binder = Binder(signature).update(self.__module)
-        self.__module.add_listener(binder)
-        return binder
-
-
-@final
-@dataclass(repr=False, frozen=True, slots=True)
-class InjectableDecorator:
-    __module: Module
-    __injectable_type: type[BaseInjectable]
-
-    def __repr__(self) -> str:
-        return f"<{self.__injectable_type.__qualname__} decorator>"
-
-    def __call__(
-        self,
-        wrapped: Callable[..., Any] = None,
-        /,
-        *,
-        on: type | Iterable[type] = None,
-    ):
-        def decorator(wp):
-            @lambda fn: fn()
-            def classes():
-                if cls := self.__get_target_class(wp):
-                    yield cls
-
-                if on is None:
-                    return
-                elif isinstance(on, type | str):
-                    yield on
-                else:
-                    yield from on
-
-            @self.__module.inject
-            @wraps(wp, updated=())
-            def factory(*args, **kwargs):
-                return wp(*args, **kwargs)
-
-            injectable = self.__injectable_type(factory)
-            self.__module.update(classes, injectable)
-
-            return wp
-
-        return decorator(wrapped) if wrapped else decorator
-
-    @staticmethod
-    def __get_target_class(wrapped: Callable[..., Any]) -> type | None:
-        if isinstance(wrapped, type):
-            return wrapped
-
-        if callable(wrapped):
-            return_type = get_annotations(wrapped, eval_str=True).get("return")
-
-            if isinstance(return_type, type):
-                return return_type
-
-        return None
